#include "minmax-chisq_fn.h"
FILE *fopene(char *filename, char *rw, char *err){
  FILE *fp;
  if((fp=fopen(filename,rw))==NULL){
    printf("%s\n",err);
    exit(1);
  }
  return(fp);
}
char get_param(int *k, char **arg, char *value){
  char opt='z';
  int i;
  
  for(i = *k; i >= 0; i--){
    if(arg[i][0]=='-') opt=arg[i][1];
    if(arg[i][2]!='\0') sprintf(value,"%s",&arg[i][2]);
    if(arg[i][2]=='\0') sprintf(value,"%s",arg[i+1]);
    if(opt!='z'){
      *k=i-1;
      break;
    }
  }
  return(opt);
}
void get_params(int n, char **arg, FILE **binfile, FILE **ibinfile, 
		int *iseed, int *nrand, int *nbinl, int *nbinu){
  char opt,value[100];
  int k;
  k=n-1;
  *binfile=stdout;
  *ibinfile=NULL;
  *iseed=42;
  *nrand=1000;
  *nbinl=2;
  *nbinu=20;
  while((opt=get_param(&k,arg,value)) != 'z'){
/*     printf("%c %s\n",opt,value); */
    if(opt=='i') *ibinfile=fopene(value,"r","infile not available");
    if(opt=='o') *binfile=fopen(value,"w");
    if(opt=='s') *iseed=atoi(value);
    if(opt=='n') *nrand=atoi(value);
    if(opt=='l') *nbinl=atoi(value);
    if(opt=='u') *nbinu=atoi(value);
  }
  return;
}
void binrandf(int nbin, int *bin){
  /* random bin generator was required. Starting with nchar bins, bins are
   * randomly chosen for joing until the right number has been found.
   *
   * required because the snippit of code
   * nunique <- 0
   * while(nunique < nbin){
   *   bin.rand <- sample(c(1:nbin),NCHAR,replace=T)
   *   nunique <- length(unique(bin.rand))
   * }
   * took too long to run when there were a large number of bins
   * translated from apr18_2006.q */
  int i,nb,l1,l2,tmp;
  for(i = 0; i < NCHAR; i++) bin[i] = i;
  nb = NCHAR;
  while(nb > nbin){
    sample(1,nbin,&l1); /* sample bins for join with replace from 0:nbin*/
    l1--;
    sample(1,nbin-1,&l2);
    if (l2 >= l1) l2++;
    if (l2 < l1){ /* l2 is larger of bin labels */
      tmp = l1;
      l1 = l2;
      l2 = tmp;
    }
    
    for(i = 0; i < NCHAR; i++) /* bins l1 and l2 become l1 */
      if (bin[i] == l2) bin[i] = l1;
    if (l2 != nb-1) /* relabel last bin l2 */
      for(i = 0; i < NCHAR; i++)  
	if (bin[i] == nb-1) bin[i] = l2;

    nb--;
  }
}
int l2ip(char c){
  switch(c){
  case 'A':
    return(alanine);
    
  case 'R':
    return(arginine);
    
  case 'N':
    return(asparagine);
    
  case 'D':
    return(aspartic);

  case 'C':
    return(cysteine);

  case 'Q':
    return(glutamine);

  case 'E':
    return(glutamic);

  case 'G':
    return(glycine);

  case 'H':
    return(histidine);

  case 'I':
    return(isoleucine);

  case 'L':
    return(leucine);

  case 'K':
    return(lysine);

  case 'M':
    return(methionine);

  case 'F':
    return(phenylalanine);

  case 'P':
    return(proline);

  case 'S':
    return(serine);

  case 'T':
    return(threonine);

  case 'W':
    return(tryptophan);

  case 'Y':
    return(tyrosine);

  case 'V':
    return(valine);

  case '-':
    return(21);
    
  case 'X':
    return(22);

  case '?':
    return(23);
    /* future: cases X and ? represent unknown aa 
     * B - asparagine or aspartic
     * Z - glutamine or glutamic
     * * - stop codon
     */
  }
  return(100);
}
void rmgap(int numsp, int *nsite, char *seqc){
  int i,j,k,sp,newnsite;

  newnsite = *nsite;
  for(i = *nsite-1; i >= 0; i--)
    for(j = 0; j < numsp; j++)
      if (seqc[i + j*(*nsite)] == '-'){
	for(k = i; k < newnsite-1; k++)
	  for(sp = 0; sp < numsp; sp++)
	    seqc[k + sp*(*nsite)] = seqc[k+1 + sp*(*nsite)];
	newnsite--;
	break;
      }

  for(j = 1; j < numsp; j++)
    for(i = 0; i < newnsite; i++)
      seqc[i + j*newnsite] = seqc[i + j*(*nsite)];

  *nsite = newnsite;
}
int rinterleave_block(FILE *infile, int numsp, int nsite, char name[][11], 
		       char *seq, int *initb){
  char c;
  int i,j,lsite,lsiteo;
  
  for(i = 0; i < numsp; i++){
    c = ignore_whitespace_seqfile(infile);
    if((*initb) == 1){
      name[i][0] = c;
      for(j = 1; j < 10; j++)name[i][j] = getc(infile);
      name[i][10] = '\0';
      c = ignore_whitespace_seqfile(infile);
    }
    seq[i*nsite] = c;
    lsite = rseqf(infile,&seq[1+i*nsite])+1;
    if(i==0) lsiteo=lsite;

    if(i > 0 && lsite != lsiteo){
      printf("seqfile: number of character states should be same for all taxa in a block\n");
      exit(1);
    }
  }
  *initb = 0;
  
  return(lsite);
}
char ignore_whitespace_seqfile(FILE *infile){
  char c;
  while((c = getc(infile)) == ' ' || c == '\n' || c == '\r' || c == '\t') ;
  if(c == EOF){
    printf("seqfile: EOF before entire sequence read");
    exit(1);
  }
  return(c);
}
void rinterleavef_nolim(FILE *infile, const int *numsp, const int *nsite, 
			char name[][11], char *seq)
{
  int lsite = 0, csite = 0, i, j, initb = 1;
  char c;

  while(csite < *nsite){
    lsite = rinterleave_block(infile,*numsp,*nsite,name,&seq[csite],&initb);
    csite += lsite;
  }
}
void letter(int numsp, int nsite, char *seq)
{
  int i, j;
  for(i = 1; i < numsp; i++)
    for(j = 0; j < nsite; j++)
      if (seq[j + i*nsite] == '.') seq[j + i*nsite]=seq[j];
}
int rseqf(FILE *fp, char *seq){
  int rsite = 0;
  char c;
  
  while((c = getc(fp)) != '\n' &&  c != EOF){
    if (c != ' ' && c != '\r' && c!= '\t'){
      *seq = c;
      seq++;
      rsite++;
    }
  }

  return(rsite);
}
void pchar_freq(int nchar, int nsite, int *s1, int *s2, double *pf)
{
  int i,j,nc=0;

  for(j = 0; j < nchar*nchar; j++) pf[j] = 0.0;

  for(i = 0; i < nsite; i++){
    if(s1[i] < nchar && s2[i] < nchar){
      pf[s1[i]+s2[i]*nchar]++;
      nc++;
    }
  }

  for(j = 0; j < nchar*nchar; j++) pf[j] /= nc;
}
void char_freq(int nchar, int numsp, int nsite, int *seq, double *freq)
{
  int i,j,nc=0;

  for(j = 0; j < nchar; j++){
    freq[j] = 0.0;
  }
  for(i = 0; i < nsite*numsp; i++){
    if(seq[i] < nchar){
      for(j = 0; j < nchar; j++){
	if (seq[i] == j) freq[j]++;
      }
      nc++;
    }
  }

  for(j = 0; j < nchar; j++){
    freq[j] /= nc;
  }
}
int pair_Fmat(int nsite, int nbin, int *s1, int *s2, double *Fxy, double *Pi){ 
  /* return 1 if Pi[i] > 1, some i */
  int i,j,site;

  for(i = 0; i < nbin; i++)
    for(j = i; j < nbin; j++){
      Fxy[i+j*nbin] = 0;
      for(site = 0; site < nsite; site++)
	if ((s1[site] == i && s2[site] == j) || 
	    (s1[site] == j && s2[site] == i) ) Fxy[i+j*nbin]++;
    }

  for(i = 0; i < nbin; i++)
    for(j = (i+1); j < nbin; j++) Fxy[i+j*nbin] /= (2*nsite);
  for(i = 0; i < nbin; i++) Fxy[i+i*nbin] /= nsite;

  for(i = 0; i < nbin; i++)
    for(j = 0; j < i; j++) Fxy[i+j*nbin] = Fxy[j+i*nbin];
  
  for(i = 0; i < nbin; i++){
    Pi[i] = 0.0;
    for(j = 0; j < nbin; j++) Pi[i] += Fxy[i+j*nbin];
    if (Pi[i] <= 0.0){
      printf("Frequency = 0\n");
      return(0);
    }
  }
  return(1);
}
int log_piFxy(int nbin, double *Fxy, double *Pi, double *q){
  int i,j,k;
  double Offdiag[400],W[400],Lam[20];

/*   printf("\n"); wmat(nbin, nbin, Fxy); printf("\n"); */
  for(i = 0; i < nbin; i++) /* Pi**(-1/2) Fxy Pi**(-1/2) */
    for(j = 0; j < nbin; j++) W[i+j*nbin] = Fxy[i+j*nbin] / (sqrt(Pi[i]*Pi[j]));
/*   printf("\n"); wmat(nbin, nbin, W); printf("\n"); */
  eigenRealSym(W, nbin, Lam, Offdiag); /* eigen decomposition */

/*   for(i = 0; i < nbin; i++) printf("%f ", Lam[i]); printf("\n"); */
  for(i = 0; i < nbin; i++) /* check eigen values > 0 and take logs */
    if (Lam[i] < 0.0){
/*       printf("negative eigenvalue\n"); */
      return(0);
    }
    else
      Lam[i] = log(Lam[i]);

  for(i = 0; i < nbin; i++) /* log = Pi**(-1/2) [W**T * Lam * W] Pi**(1/2) */
    for(j = 0; j < nbin; j++){
      q[i+j*nbin] = 0;
      for(k = 0; k < nbin; k++) 
	q[i+j*nbin] += W[k+i*nbin]*Lam[k]*W[k+j*nbin];
      q[i+j*nbin] *= sqrt(Pi[j]/Pi[i]);
    }

  /* printf("log(Pi**(-1) F_xy)\n");wmat(nbin, nbin, q); printf("\n"); */
  return(1); /* successful return */
}
double gtr_dist(int nbin, double *q, double *Pi){
  int i;
  double d = 0.0;
  for(i = 0; i < nbin; i++) d -= Pi[i] * q[i+i*nbin];
  return(d);
}
double gtrpair_Qdist(int nsite, int nbin, int *s1, int *s2, double *q){
  /* return -1 if Pi[i] = 0 Lam[i] = 0 or d[i] = 0 */
  int i,j;
  double d, Fxy[NCHAR*NCHAR], Pi[NCHAR];

  /* frequencies for pair and Fxy matrix */
  if (!pair_Fmat(nsite, nbin, s1, s2, Fxy, Pi)) return(-1.0); /* Fxy, Pi */
  if (!log_piFxy(nbin, Fxy, Pi, q)) return(-1.0); /* log(Pi**(-1) F_xy) for pair */
  if ((d = gtr_dist(nbin, q, Pi)) <= 0.0) return(-1.0); /* distance for pair */

  for(j = 0; j < nbin; j++)/* Q estimate for pair */
    for(i = 0; i < nbin; i++)
	q[i+j*nbin] /= d;
/*   wmat(nbin,nbin,q); printf("\n"); */
  return(d);
}
int gtrpair_Qdists(int numsp, int nsite, int nbin, int nel, int *sb, double *d, 
		   double *q){
  /* return 0 if Pi[i] = 0 or for some pair Lam[i] = 0 or d[i] = 0 */
  int i,j,np = 0;

  for(i = 0; i < numsp; i++)
    for(j = (i+1); j < numsp; j++){
      d[i+j*numsp] = gtrpair_Qdist(nsite,nbin,&sb[i*nsite],&sb[j*nsite],&q[np*nel]);
      if (d[i+j*numsp] < 0.0) return(0);
 /*      printf("%i %i %f\n", i, j, d[i+j*numsp]); */
      np++;
    }
/*    exit(0); */

  for(i = 0; i < numsp; i++)
    for(j = 0; j < i; j++) d[i+j*numsp] = d[j+i*numsp];
  for(i = 0; i < numsp; i++) d[i+i*numsp] = 0.0;
  return(1);
}
double ls_q(int nel, double *q1, double *q2){
  int i;
  double ss = 0.0;
  for(i = 0; i < nel; i++) ss += (q1[i] - q2[i])*(q1[i] - q2[i]);
  return(ss);
}
double ls_gtr(int numsp, int nsite, int nbin, int *sb, double *dmat){
  /* return 1.0e+20 if Pi[i] = 0 or for some pair Lam[i] = 0 or d[i] = 0 */
  int nel,npair,i,j;
  double ss = 0.0, *q;

/*   FILE *q_optfile; */
  
  /*  nel = nbin*nbin - nbin; */
  nel = nbin*nbin;
  npair = ((numsp-1)*numsp)/2;
  q = cvec(npair*nbin*nbin);
  if(!gtrpair_Qdists(numsp, nsite, nbin, nel, sb, dmat, q)) ss = 1.0e+20;

/*   q_optfile = fopen("q_optfile_initial", "w"); */
/*   for(i = 0; i < nel; i++) */
/*     for(j = 0; j < npair; j++) fprintf(q_optfile, "%e ", q[i+j*nel]); */
/*   fclose(q_optfile); */
/*   exit(0); */

  if (ss < 1.0){
    for(i = 0; i < npair; i++) 
      for(j = (i+1); j < npair; j++) ss += ls_q(nel, &q[i*nel], &q[j*nel]);
  }
  free(q);
  return(ss);
}
int nel_bin(int binl, int *bin){
  int nel = 0, i;
  for(i = 0; i < NCHAR; i++) 
    if (bin[i] == binl) nel++;
  return(nel);
}
void bin_rearrangement(int nbin, int *bin, int *elp, int *bap, int *newbin)
{
  /* all rearrarangements of one el from a bin to another.   
   * i/o (reverse communication):               
   * el - current el in bin for possible move     
   * ba - bin for placement of el          
   *                                    
   * el and ba are updated; input for next bin rearrangement   
   * defaults for new set of rearrangements          
   * el >= nchar if no rearrangments possible      
   *                                             
   * routine attempts to update ba if possible, o/w el   
   */
  int i,done,el,ba;
 
  el = *elp; ba = *bap;
  for(i = 0; i < NCHAR; i++) newbin[i] = bin[i];
  done = 0;

  if (nel_bin(bin[el], bin) == 1){ /* make sure bin has more than one element */
    while(el < NCHAR && nel_bin(bin[el], bin) == 1) el++;
    ba = -1; /* will be updated  */
  }

  while(!done && el < NCHAR){
    ba = ba+1;
    if(ba == bin[el]) ba = ba+1; /* find next alt bin for placement */ 
    if(ba < nbin){
      newbin[el] = ba;
      done = 1;
    }
    else{ /* no alt bin available   */            
      el++; /* next el in bin with more than one el */                             
      while(el < NCHAR && nel_bin(bin[el], bin) == 1) el++;
      ba = -1; /* will be updated */
    }
  }
  /*  pbin2bins(nbin, newbin); */

  *elp = el;
  *bap = ba;
}
void s2binm(int numsp, int nsite, int *seq, int *bin, int *seqb){
  int i;
  for(i = 0; i < nsite*numsp; i++) seqb[i] = bin[seq[i]];
}
double min_eig_pair_gtr(int nsite, int nbin, int *s1, int *s2){
  int i,j;
  double min_eig,Offdiag[NCHAR*NCHAR],W[NCHAR*NCHAR],Lam[NCHAR],Fxy[NCHAR*NCHAR],Pi[NCHAR];
  
  if (!pair_Fmat(nsite,nbin,s1,s2,Fxy,Pi)) return(-1.0e+20); 
  for(i = 0; i < nbin; i++) /* Pi**(-1/2) Fxy Pi**(-1/2) */
    for(j = 0; j < nbin; j++) W[i+j*nbin] = Fxy[i+j*nbin] / (sqrt(Pi[i]*Pi[j]));
  eigenRealSym(W, nbin, Lam, Offdiag); /* eigen decomposition */
  min_eig = Lam[0];
  for(i = 1; i < nbin; i++) if(Lam[i] < min_eig) min_eig = Lam[i];
  return(min_eig);
}
double min_eig_gtr(int numsp, int nsite, int nbin, int *s){
  int i,j;
  double min_eigt,min_eig = 1.0e+20;
/*   for(i = 0; i < nsite*numsp; i++) printf("%i ", s[i]); printf("\n"); */
  for(i = 0; i < numsp; i++)
    for(j = (i+1); j < numsp; j++){
      min_eigt = min_eig_pair_gtr(nsite,nbin,&s[i*nsite],&s[j*nsite]);
      if (min_eigt < min_eig) min_eig = min_eigt;
    }
  return(min_eig);
}
double maxmin_eig_gtr(int numsp, int nsite, int nbin, int *seq, int *bin, int *seqb){
  int i,el,ba,nb=0,bint[NCHAR];
  double crit,crit_new;
  crit = -1.0e+20;
  for(i = 0; i < NCHAR; i++) bint[i] = bin[i];
  el = 0; ba = -1;

  while(crit < 0.0 && el < NCHAR){
    s2binm(numsp, nsite, seq, bint, seqb);
    crit_new = min_eig_gtr(numsp, nsite, nbin, seqb);
    if (crit_new > crit){
      crit = crit_new;
      for(i = 0; i < NCHAR; i++) bin[i] = bint[i];
      el = 0; ba = -1;
      printf("Minimum Eigenvalue: %f\n", crit_new);
    }
    bin_rearrangement(nbin, bin, &el, &ba, bint);
    nb++;
  }

  printf("%i bin rearrangements were tried\n",nb);
  return(crit);
}
double best_bin_gtr(int numsp, int nsite, int nbin, int *seq, int *bin, 
		    double *dmat){
  int i,el,ba,nb=0,*seqb,bint[NCHAR];
  double crit,crit_new;

  seqb = civec(nsite*numsp);
  if (maxmin_eig_gtr(numsp,nsite,nbin,seq,bin,seqb) < 0.0) return(1.0e+20);
  crit = ls_gtr(numsp, nsite, nbin, seqb, dmat);
  printf("Initial criterion: %f\n", crit);
  el = 0; ba = -1;
  while(el < NCHAR){
    bin_rearrangement(nbin, bin, &el, &ba, bint);
/*     for(i = 0; i < NCHAR; i++) printf("%i ", bint[i]); printf("\n"); */
    s2binm(numsp, nsite, seq, bint, seqb);
    crit_new = ls_gtr(numsp, nsite, nbin, seqb, dmat);
/*     printf("Criterion: %f\n", crit_new); */
    if (crit_new < crit){
      crit = crit_new;
      printf("Criterion: %f\n", crit_new);
      el = 0; ba = -1;
      for(i = 0; i < NCHAR; i++) bin[i] = bint[i]; 
    }
    nb++;
  }
  printf("%i bin rearrangements were tried\n",nb);
  s2binm(numsp, nsite, seq, bin, seqb); /* distance matrix, crit for final bin */
  crit = ls_gtr(numsp, nsite, nbin, seqb, dmat);

  free(seqb);
  return(crit);
}
int num_bin(int nbin, int *bin){
  int nb = 0, i,j;
  for(i = 0; i < nbin; i++)
    for(j = 0; j < NCHAR; j++) 
      if (bin[j] == i){
	nb++;
	break;
      }
  return(nb);
}
void opt_q(int numsp, int nsite, int nbin, int *seq, int *bin, double *q){
  int nel,*sb;
  double *dmat;
  sb = civec(numsp*nsite);
  dmat = cvec(numsp*numsp);
  s2binm(numsp, nsite, seq, bin, sb);
  nel = nbin*nbin;
  gtrpair_Qdists(numsp, nsite, nbin, nel, sb, dmat, q);
  free(sb);
  free(dmat);
}
 double  best_bin_gtr_rand(int numsp, int nsite, int nbin, int nrand, int *seq, 
			   int *bin, double *dmat){
   FILE *q_optfile; 
   int isamp,i,j,npair,nel,bint[NCHAR];
   double crit,crit_new,*q_opt,*dmatt;

   npair = (numsp*(numsp-1))/2;
   q_opt = cvec(npair*nbin*nbin);
   dmatt = cvec(numsp*numsp);

   for(isamp = 1; isamp <= nrand; isamp++){
     sample(NCHAR, nbin, bint); /* 20 integers from 1:nbin */
     for(i = 0; i < NCHAR; i++) bint[i]--; 
     while(num_bin(nbin, bint) < nbin){ /* no reduction in number of bins */
       sample(NCHAR, nbin, bint); 
       for(i = 0; i < NCHAR; i++) bint[i]--; 
     }
     for(i = 0; i < NCHAR; i++) printf("%i ", bint[i]); printf("\n");
     crit_new = best_bin_gtr(numsp,nsite,nbin,seq,bint,dmatt);
     if (isamp == 1){
       for(i = 0; i < NCHAR; i++) bin[i] = bint[i];
       for(i = 0; i < numsp*numsp; i++) dmat[i] = dmatt[i];
       crit = crit_new;
     }
     if (isamp > 1 &&  crit_new < crit){
	 for(i = 0; i < NCHAR; i++) bin[i] = bint[i];
	 for(i = 0; i < numsp*numsp; i++) dmat[i] = dmatt[i];
	 crit = crit_new;
     }
     
   }
   printf("Criterion: %f\n", crit);
   opt_q(numsp, nsite, nbin, seq, bin, q_opt);
   q_optfile = fopen("q_optfile", "w");
   nel = nbin*nbin;
   for(i = 0; i < nel; i++)
     for(j = 0; j < npair; j++) fprintf(q_optfile, "%e ", q_opt[i+j*nel]);
   fclose(q_optfile);

   free(q_opt); free(dmatt);
   return(crit);
 }
void get_nsitev(int numsp, int nsite, int *seq, int *nsitev){
  int i,j;
  for(j = 0; j < numsp; j++){
    nsitev[j] = 0;
    for(i = 0; i < nsite; i++)
      if (seq[i+j*nsite] < 20) nsitev[j]++;
  }
}
void create_fmat(int ntaxa, int nsite, int *seq, double *fmat, double *fr){
  /* ith column of fmat contains frequencies of amino acids for ith taxa
   * ith entry of fr, the overall frequencies
   */
  int i;
  for(i = 0; i < ntaxa; i++) char_freq(NCHAR,1,nsite,&seq[i*nsite],&fmat[i*NCHAR]);
  char_freq(NCHAR,ntaxa,nsite,seq,fr);
}
void create_pfarray(int ntaxa, int nsite, int *seq, double *pfa){
  int np = 0,i,j;
  for(i = 0; i < ntaxa; i++) 
    for(j = (i+1); j < ntaxa; j++){
      pchar_freq(NCHAR,nsite,&seq[i*nsite],&seq[j*nsite],
		 &pfa[np*NCHAR*NCHAR]);
      np++;
    }
}
void create_pfarray_bin(int nbin, int *bin, double *pf, double *pf_bin){
  int i, j, k, l;
  for(i = 0; i < nbin*nbin; i++) pf_bin[i] = 0.0;

  for(i = 0; i < nbin; i++)
    for(j = 0; j < nbin; j++)
      for(k = 0; k < 20; k++)
	for(l = 0; l < 20; l++)
	  if (bin[k] == i && bin[l] == j) pf_bin[i+j*nbin] += pf[k+l*20];
  
}
double chisq_crit(int ntaxa, int *nsitev, int nbin, int *bin, double *fmat, 
		  double *fr, double *pfa){
  /* average chi-squared criterion; not adjusting for number of sites */
  int i,j,k;
  double frb[NCHAR],frt,chisq=0.0;
  
  for(i = 0; i < nbin; i++){
    frb[i] = 0.0;
    for(j = 0; j < NCHAR; j++)
      if(bin[j] == i) frb[i] += fr[j];
  }

  for(k = 0; k < ntaxa; k++){
    for(i = 0; i < nbin; i++){
      frt = 0.0;
      for(j = 0; j < NCHAR; j++)
	if(bin[j] == i) frt += fmat[j+k*NCHAR];
      chisq += (frt - frb[i])*(frt - frb[i])/frb[i];
    }
  }
  return(chisq);
}
double chisq_max_crit(int ntaxa, int *nsitev, int nbin, int *bin, double *fmat, 
		      double *fr, double *pfa){
  /* maximum chi-squared criterion */
  int i,j,k;
  double frb[NCHAR],frt,chisq,max_chisq=0.0;
  
  for(i = 0; i < nbin; i++){
    frb[i] = 0.0;
    for(j = 0; j < NCHAR; j++)
      if(bin[j] == i) frb[i] += fr[j];
  }

  for(k = 0; k < ntaxa; k++){
    chisq = 0.0;
    for(i = 0; i < nbin; i++){
      frt = 0.0;
      for(j = 0; j < NCHAR; j++)
	if(bin[j] == i) frt += fmat[j+k*NCHAR];
      chisq += (frt - frb[i])*(frt - frb[i])/frb[i];
    }
    chisq *= nsitev[k];
/*     printf("%6.2f%\n",100*(1-pchisq(chisq,19))); */
    if (chisq > max_chisq) max_chisq = chisq;
  }
  return(max_chisq);
}
void prfmat(int ntaxa, int nbin, int *bin, double *fmat, double *fr){
  /* testing routine */
  int i,j,k;
  double frb[NCHAR],frt;

  for(i = 0; i < nbin; i++){
    frb[i] = 0.0;
    for(j = 0; j < NCHAR; j++)
      if(bin[j] == i) frb[i] += fr[j];
  }

  for(k = 0; k < ntaxa; k++){
    for(i = 0; i < nbin; i++){
      frt = 0.0;
      for(j = 0; j < NCHAR; j++)
	if(bin[j] == i) frt += fmat[j+k*NCHAR];
      printf("%f ", frt);
    }
    printf("\n");
  }
}
double percent_crit(int ntaxa, int *nsitev, int nbin, int *bin, double *fmat, 
		    double *fr, double *pfa){
  int i,j,k;
  double frb[NCHAR],frt,chisq=0.0;
  
  for(i = 0; i < nbin; i++){
    frb[i] = 0.0;
    for(j = 0; j < NCHAR; j++)
      if(bin[j] == i) frb[i] += fr[j];
  }

  for(k = 0; k < ntaxa; k++){
    for(i = 0; i < nbin; i++){
      frt = 0.0;
      for(j = 0; j < NCHAR; j++)
	if(bin[j] == i) frt += fmat[j+k*NCHAR];
      chisq += (frt - frb[i])*(frt - frb[i])/(frb[i]*frb[i]);
    }
  }
  return(chisq);
}
double entropy_crit(int ntaxa, int *nsitev, int nbin, int *bin, double *fmat, 
		    double *fr, double *pfa){
  int i,j;
  double frb[NCHAR],entrop=0.0;
  
  for(i = 0; i < nbin; i++){
    frb[i] = 0.0;
    for(j = 0; j < NCHAR; j++)
      if(bin[j] == i) frb[i] += fr[j];
    entrop += frb[i]*log(frb[i]);
  }

  return(entrop);
}
double entropy_ave_crit(int ntaxa, int *nsitev, int nbin, int *bin, 
			double *fmat, double *fr, double *pfa){
  int i,j,k;
  double frt,entrop=0.0;

  for(k = 0; k < ntaxa; k++){
    for(i = 0; i < nbin; i++){
      frt = 0.0;
      for(j = 0; j < NCHAR; j++)
	if(bin[j] == i) frt += fmat[j+k*NCHAR];
      entrop += frt*log(frt);
    }
  }
  return(entrop);
}
double match_crit_pair(int nbin, int *bin, double *pf){
  /* It was recognized later that only first nbin-1 elements were required. 
   * Only necessary changes have been done to reflect that.
   * Some unnecessary compuation is being done */
  int i,j;
  double pf_bin[400],f1_bin[20],f2_bin[20],d[20],v[400],u[20],ss2;

  for(i = 0; i < nbin; i++) f1_bin[i] = 0.0;
  for(i = 0; i < nbin; i++) f2_bin[i] = 0.0;

  create_pfarray_bin(nbin,bin,pf,pf_bin);

  for(i = 0; i < nbin; i++)
    for(j = 0; j < nbin; j++) f1_bin[i] += pf_bin[i+j*nbin];

  for(j = 0; j < nbin; j++) 
    for(i = 0; i < nbin; i++) f2_bin[j] += pf_bin[i+j*nbin];

  for(i = 0; i < nbin; i++) d[i] = f1_bin[i] - f2_bin[i];

/*   for(i = 0; i < nbin; i++){  */
/*     for(j = 0; j < nbin; j++) printf("%.16e ", pf_bin[i+j*nbin]); */
/*     printf("\n"); */
/*   } */
/*   printf("\n"); */

  for(i = 0; i < (nbin-1); i++)
    for(j = 0; j < (nbin-1); j++) 
      if (i == j)
	v[i+i*(nbin-1)] = f1_bin[i] + f2_bin[i] - 2.0*pf_bin[i+i*nbin];
      else
	v[i+j*(nbin-1)] = -1.0*(pf_bin[i+j*nbin] + pf_bin[j+i*nbin]);


/*   for(i = 0; i < nbin; i++){  */
/*     for(j = 0; j < nbin; j++) printf("%.16e ", v[i+j*nbin]); */
/*     printf("\n"); */
/*   } */

  if(chol(nbin-1,v)){
    printf("Difficulty computing Cholesky decomposition of variance matrix\n");
    exit(0);
  }
  for(i = 0; i < nbin; i++) u[i] = d[i];
  cholh(nbin-1,v,u);

  ss2 = 0.0;
  for(i = 0; i < nbin-1; i++) ss2  += d[i]*u[i];
  return(ss2);
}
double match_crit(int ntaxa, int *nsitev, int nbin, int *bin, double *fmat, 
		  double *fr, double *pfa){
  int np = 0,i,j;
  double ssmax = 0.0,ss;
  for(i = 0; i < ntaxa; i++)
    for(j = (i+1); j < ntaxa; j++){
      ss = match_crit_pair(nbin,bin,&pfa[400*np]);
      np++;
      if (ss > ssmax) ssmax = ss;
    }
  return(ssmax);
}
double bowker_crit_pair(int nbin, int *bin, double *pf){
  int i,j;
  double ss,pf_bin[400];

  create_pfarray_bin(nbin,bin,pf,pf_bin);
  ss = 0.0;
  for(i = 0; i < nbin; i++)
    for(j = (i+1); j < nbin; j++){
      if ((pf_bin[i+j*nbin] + pf_bin[j+i*nbin]) > 0.0){
	ss += (pf_bin[i+j*nbin] - pf_bin[j+i*nbin]) *
	  (pf_bin[i+j*nbin] - pf_bin[j+i*nbin]) / 
	  (pf_bin[i+j*nbin] + pf_bin[j+i*nbin]);
      }
      else{
	ss = 1.0e+20;
	return(ss);
      }
    }
  return(ss);
}
double bowker_crit(int ntaxa, int *nsitev, int nbin, int *bin, double *fmat, 
		   double *fr, double *pfa){
  int np = 0,i,j;
  double ssmax = 0.0,ss;
   for(i = 0; i < ntaxa; i++)
    for(j = (i+1); j < ntaxa; j++){
      ss = bowker_crit_pair(nbin,bin,&pfa[400*np]);
      np++;
      if (ss > ssmax) ssmax = ss;
    }
  return(ssmax);
}
double best_bin_chisq(int ntaxa, int *nsitev, int nbin, int *bin, double *fmat, 
		      double *fr, double *pfa, int crit_id){
  /* because some criterion may be exactly the same for two sets of bins, 
   * we only update the bins when a decrease less than eps has been found. 
   */
  int i,el,ba,nb=0,bint[NCHAR];
  double crit,crit_new=1.0e+20,eps=1.0e-8;
  double (*crit_fn)(int,int*,int,int *,double *,double *,double *);
  
  if (crit_id == 0) crit_fn = chisq_crit;
  if (crit_id == 1) crit_fn = chisq_max_crit;
  if (crit_id == 2) crit_fn = percent_crit;
  if (crit_id == 3) crit_fn = entropy_crit;
  if (crit_id == 4) crit_fn = entropy_ave_crit;
  if (crit_id == 5) crit_fn = match_crit;
  if (crit_id == 6) crit_fn = bowker_crit;

  crit = (*crit_fn)(ntaxa,nsitev,nbin,bin,fmat,fr,pfa);
/*   printf("Initial criterion: %f\n", crit); */
/*   wirow(stdout,20,bin,1); */
  el = 0; ba = -1;
  while(el < NCHAR){
    bin_rearrangement(nbin, bin, &el, &ba, bint);
    if (el < NCHAR)
      crit_new = (*crit_fn)(ntaxa,nsitev,nbin,bint,fmat,fr,pfa);    
    if (crit_new < crit-eps){ /* dangerous */
      crit = crit_new;
/*       wirow(stdout,NCHAR,bin,1); */
/*       printf("Criterion: %f\n", crit_new); */
      el = 0; ba = -1;
      for(i = 0; i < NCHAR; i++) bin[i] = bint[i]; 
    }
    nb++;
  }
/*   printf("%i bin rearrangements were tried\n",nb); */
  crit = (*crit_fn)(ntaxa,nsitev,nbin,bin,fmat,fr,pfa);
/*   printf("Final Criterion %f\n", crit); */

  return(crit);
}
int eigenRealSym(double A[], int n, double Root[], double Offdiag[])
{
/* This finds the eigen solution of a real symmetrical matrix A[n*n].  In return, 
   A has the right vectors and Root has the eigenvalues. work[n] is the working space.
   The matrix is first reduced to a tridiagonal matrix using HouseholderRealSym(), 
   and then using the QL algorithm with implicit shifts.  

   Adapted from routine tqli in Numerical Recipes in C, with reference to LAPACK
   Ziheng Yang, 23 May 2001
*/
   int status=0;
   HouseholderRealSym(A, n, Root, Offdiag);
   status=EigenTridagQLImplicit(Root, Offdiag, n, A);
   EigenSort(Root, A, n);

   return(status);
}
void EigenSort(double d[], double U[], int n)
{
/* this sorts the eigen values d[] and rearrange the (right) eigen vectors U[]
*/
   int k,j,i;
   double p;

   for (i=0;i<n-1;i++) {
      p=d[k=i];
      for (j=i+1;j<n;j++)
         if (d[j] >= p) p=d[k=j];
      if (k != i) {
         d[k]=d[i];
         d[i]=p;
         for (j=0;j<n;j++) {
            p=U[j*n+i];
            U[j*n+i]=U[j*n+k];
            U[j*n+k]=p;
         }
      }
   }
}
void HouseholderRealSym(double a[], int n, double d[], double e[])
{
/* This uses HouseholderRealSym transformation to reduce a real symmetrical matrix 
   a[n*n] into a tridiagonal matrix represented by d and e.
   d[] is the diagonal (eigends), and e[] the off-diagonal.
*/
   int m,k,j,i;
   double scale,hh,h,g,f;

   for (i=n-1;i>=1;i--) {
      m=i-1;
      h=scale=0;
      if (m > 0) {
         for (k=0;k<=m;k++)
            scale += fabs(a[i*n+k]);
         if (scale == 0)
            e[i]=a[i*n+m];
         else {
            for (k=0;k<=m;k++) {
               a[i*n+k] /= scale;
               h += a[i*n+k]*a[i*n+k];
            }
            f=a[i*n+m];
            g=(f >= 0 ? -sqrt(h) : sqrt(h));
            e[i]=scale*g;
            h -= f*g;
            a[i*n+m]=f-g;
            f=0;
            for (j=0;j<=m;j++) {
               a[j*n+i]=a[i*n+j]/h;
               g=0;
               for (k=0;k<=j;k++)
                  g += a[j*n+k]*a[i*n+k];
               for (k=j+1;k<=m;k++)
                  g += a[k*n+j]*a[i*n+k];
               e[j]=g/h;
               f += e[j]*a[i*n+j];
            }
            hh=f/(h*2);
            for (j=0;j<=m;j++) {
               f=a[i*n+j];
               e[j]=g=e[j]-hh*f;
               for (k=0;k<=j;k++)
                  a[j*n+k] -= (f*e[k]+g*a[i*n+k]);
            }
         }
      } 
      else
         e[i]=a[i*n+m];
      d[i]=h;
   }
   d[0]=e[0]=0;

   /* Get eigenvectors */
   for (i=0;i<n;i++) {
      m=i-1;
      if (d[i]) {
         for (j=0;j<=m;j++) {
            g=0;
            for (k=0;k<=m;k++)
               g += a[i*n+k]*a[k*n+j];
            for (k=0;k<=m;k++)
               a[k*n+j] -= g*a[k*n+i];
         }
      }
      d[i]=a[i*n+i];
      a[i*n+i]=1;
      for (j=0;j<=m;j++) a[j*n+i]=a[i*n+j]=0;
   }
}
int EigenTridagQLImplicit(double d[], double e[], int n, double z[])
{
/* This finds the eigen solution of a tridiagonal matrix represented by d and e.  
   d[] is the diagonal (eigenvalues), e[] is the off-diagonal
   z[n*n]: as input should have the identity matrix to get the eigen solution of the 
   tridiagonal matrix, or the output from HouseholderRealSym() to get the 
   eigen solution to the original real symmetric matrix.
   z[n*n]: has the orthogonal matrix as output

   Adapted from routine tqli in Numerical Recipes in C, with reference to
   LAPACK fortran code.
   Ziheng Yang, May 2001
*/
   int m,j,iter,niter=30, status=0, i,k;
   double s,r,p,g,f,dd,c,b, aa,bb;

   for (i=1;i<n;i++) e[i-1]=e[i];  e[n-1]=0;
   for (j=0;j<n;j++) {
      iter=0;
      do {
         for (m=j;m<n-1;m++) {
            dd=fabs(d[m])+fabs(d[m+1]);
            if (fabs(e[m])+dd == dd) break;  /* ??? */
         }
         if (m != j) {
            if (iter++ == niter) {
               status=-1;
               break;
            }
            g=(d[j+1]-d[j])/(2*e[j]);

            /* r=pythag(g,1); */

            if((aa=fabs(g))>1)  r=aa*sqrt(1+1/(g*g));
            else                r=sqrt(1+g*g);

            g=d[m]-d[j]+e[j]/(g+SIGN(r,g));
            s=c=1;
            p=0;
            for (i=m-1;i>=j;i--) {
               f=s*e[i];
               b=c*e[i];

               /*  r=pythag(f,g);  */
               aa=fabs(f); bb=fabs(g);
               if(aa>bb)       { bb/=aa;  r=aa*sqrt(1+bb*bb); }
               else if(bb==0)             r=0;
               else            { aa/=bb;  r=bb*sqrt(1+aa*aa); }

               e[i+1]=r;
               if (r == 0) {
                  d[i+1] -= p;
                  e[m]=0;
                  break;
               }
               s=f/r;
               c=g/r;
               g=d[i+1]-p;
               r=(d[i]-g)*s+2*c*b;
               d[i+1]=g+(p=s*r);
               g=c*r-b;
               for (k=0;k<n;k++) {
                  f=z[k*n+i+1];
                  z[k*n+i+1]=s*z[k*n+i]+c*f;
                  z[k*n+i]=c*z[k*n+i]-s*f;
               }
            }
            if (r == 0 && i >= j) continue;
            d[j]-=p; e[j]=g; e[m]=0;
         }
      } while (m != j);
   }
   return(status);
}
double pchisq(double x, int df)
{
  /* IN:
   * x - the value for with P(X <= x) is required
   * df - X above is a chi-squared random variable w/ df degrees of freedom 
   *
   * RETURN:
   * P(X <= x) */
  int i;
  double f, dchi;

  if (df%2 == 0){
    dchi = 0.50*exp(-0.50*x);
    f = 1.0 - 2.0*dchi;
    for(i = 4; i <= df; i+=2){
      dchi *= x/((double) i - 2);
      f -= 2.0*dchi;
    }
  }
  else{
    f = 2.0*pnorm(sqrt(x),0.0,1.0)-1.0;
    dchi = exp(-0.5*x)/sqrt(2.0*PI*x);
    for(i = 3; i <= df; i+=2){
      dchi *= x/((double) i - 2);
      f -= 2.0*dchi;
    }
  }

  return(f);
}
void sample(int n, int m, int *isamp){ /* n integers from 1:m */
  double uvar;
  int i;
  if (m == 1){
    *isamp = 1;
    return;
  }
  while(n-- > 0){
    uvar = urand();
    for(i = 1; i <= m; i++){
      if (uvar <= ((double) i)/((double) m)){
	*isamp = i;
	isamp++;
	break;
      }
    }
  }
}
double *cvec(int n)
{
  double *x;

  x=(double *) malloc((size_t) n*sizeof(double));

  return(x);
}
int *civec(int n)
{
  int *x;

  x = (int *) malloc((size_t) n*sizeof(int));

  return(x);
}
int chol(int m, double *l)
{
  int i = 0, ii = 0, j, ij, ik, im;

  for(i = 0, ii = 0, im = 0; i < m; i++, ii += (m+1), im += m){
    for(j = 0, ij = i;  j < i;  j++, ij += m)
      l[ij] = (l[ij] - dot(j, l+i, m, l+j, m))/l[j + j*m];
    for(ik = i; ik < im; ik += m) 
      l[ii] -= l[ik]*l[ik];
    if (l[ii] <= 0.0) 
      return(1);
    l[ii] = sqrt(l[ii]);
  }
  return(0);
}
void cholh(int m, double *l, double *b)
{
  int i;
  /* solution of l x = b */
  for(i = 0; i < m; i++)
    b[i] = (b[i] - dot(i, l+i, m, b, 1))/l[i + i*m];
}
double pnorm(double x, double a, double b)
{
  /* IN:
   * x - the value for with P(X <= x) is required
   * a - the mean of the normal distribution of X
   * b - the standard deviation of the normal distribution
   *
   * RETURN:
   * P(X <= x) */
  double rt2 = 1.41421356237309515, xhi = 2.66e+1;

  double z, xt,t, y;
  
  z = (x - a)/(rt2 * b);
  xt = fabs(z);
  if (xt >= xhi){
    return((z < 0.0)?0.0:1.0);
  }
  t = 1.0 - 7.5/(xt + 3.75);
  y = (((((((((((((((+3.328130055126039e-10
		     *t-5.718639670776992e-10)*t-4.066088879757269e-9)
		   *t+7.532536116142436e-9)*t+3.026547320064576e-8)
		 *t-7.043998994397452e-8)*t-1.822565715362025e-7)
	       *t+6.575825478226343e-7)*t+7.478317101785790e-7)
	     *t-6.182369348098529e-6)*t+3.584014089915968e-6)
	   *t+4.789838226695987e-5)*t-1.524627476123466e-4)
	 *t-2.553523453642242e-5)*t+1.802962431316418e-3)
       *t-8.220621168415435e-3)*t+2.414322397093253e-2;
  y = (((((y*t-5.480232669380236e-2)*t+1.026043120322792e-1)
	 *t-1.635718955239687e-1)*t+2.260080669166197e-1)
       *t-2.734219314954260e-1)*t + 1.455897212750385e-1;
  return((z > 0.0)?1.0 - 0.5*exp(-xt*xt)*y:0.5*exp(-xt*xt)*y);
}
double urand(void){
  return(drand48());  
/*   return(((double) rand())/(RAND_MAX+1.0)); /\* windows *\/ */
}
double dot(int n, double *u, int incu, double *v, int incv)
{
  double f;

  for(f = 0.0; n-- > 0; u += incu, v += incv)
    f += (*u) * (*v);

  return(f);
}
